from collections import namedtuple

import numpy as np
import periodictable
from scipy import constants

__all__ = [
    "speed_of_light",
    "c",
    "epsilon_0",
    "ep_0",
    "mu_0",
    "ke",
    "km",
    "h",
    "hbar",
    "elementary_charge",
    "e",
    "electron_mass",
    "m_e",
    "electron_mass_energy",
    "mc",
    "electron_radius",
    "r_0",
    "proton_mass",
    "m_p",
    "neutron_mass",
    "m_n",
    "atomic_constant_mass",
    "m_u",
    "u",
    "electron",
    "positron",
    "proton",
    "antiproton",
    "neutron",
    "antineutron",
    "Element",
    "Particle",
]

# Constants
c = speed_of_light = constants.c

ep_0 = epsilon_0 = constants.epsilon_0
mu_0 = constants.mu_0
h = constants.h
hbar = constants.hbar

ke = 1 / (4 * np.pi * ep_0)
km = mu_0 / (4 * np.pi)

e = elementary_charge = constants.e

m_e = electron_mass = constants.m_e
mc = electron_mass_energy = constants.physical_constants["electron mass energy equivalent in MeV"][0]
r_0 = electron_radius = constants.physical_constants["classical electron radius"][0]
m_p = proton_mass = constants.m_p
m_n = neutron_mass = constants.m_n
u = m_u = atomic_constant_mass = constants.physical_constants["atomic mass constant"][0]

Element = namedtuple("Element", ["name", "symbol", "mass", "charge"])

electron = Element(name="electron", symbol="e", mass=m_e, charge=-e)
positron = Element(name="positron", symbol="e+", mass=m_e, charge=e)
proton = Element(name="proton", symbol="p", mass=m_p, charge=e)
antiproton = Element(name="antiproton", symbol="p-", mass=m_p, charge=-e)
neutron = Element(name="neutron", symbol="n", mass=m_n, charge=0)
antineutron = Element(name="antineutron", symbol="n", mass=m_n, charge=0)
Particle = Element

# Get mass of each element from periodictable
elements = periodictable.core.default_table()
__all__ += periodictable.core.define_elements(elements, globals())



"""
Relativictic Difference Scheme Particle-in-Cell code (REDPIC) init file.

"""

from .accelerator import *
from .beam import *
from .constants import *
from .solver import *

__version__ = "0.7.15"
__doc__ = """Relativictic Difference Scheme Particle-in-Cell code (REDPIC)"""

__all__ = constants.__all__ + beam.__all__ + solver.__all__ + accelerator.__all__



import numpy as np
import pandas as pd
from numba import jit, prange
from scipy import misc

from .accelerator import *  # pylint: disable=W0614
from .beam import *  # pylint: disable=W0614
from .constants import *  # pylint: disable=W0614

__all__ = ["Simulation", "Sim"]


def get_field_accelerator(
    acc: Accelerator, type: str, x: np.array, y: np.array, z: np.array
) -> (np.array, np.array, np.array):
    """Get an electric or magnetic field at a specific location in the accelerator"""
    assert type in ("E", "B"), "Check type field! (E or B)"

    dz = acc.dz
    offset_x = acc.Dx(z)
    offset_xp = acc.Dxp(z)
    offset_y = acc.Dy(z)
    offset_yp = acc.Dyp(z)
    x_corr = x - offset_x
    y_corr = y - offset_y
    r_corr = np.sqrt(x_corr * x_corr + y_corr * y_corr)

    if type == "E":
        Ez = acc.Ez(z)
        dEzdz = acc.dEzdz(z)
        d2Ezdz2 = misc.derivative(acc.dEzdz, z, dx=dz, n=1)
        Ez = Ez - d2Ezdz2 * r_corr * r_corr / 4  # row remainder
        Ex = -dEzdz * x_corr / 2 + Ez * offset_xp  # row remainder
        Ey = -dEzdz * y_corr / 2 + Ez * offset_yp  # row remainder
        return Ex, Ey, Ez
    if type == "B":
        Bx = acc.Bx(z)
        By = acc.By(z)
        Bz = acc.Bz(z)
        dBzdz = acc.dBzdz(z)
        d2Bzdz2 = misc.derivative(acc.dBzdz, z, dx=dz, n=1)
        Gz = acc.Gz(z)
        Bz = Bz - d2Bzdz2 * r_corr * r_corr / 4  # row remainder
        Bx = Bx + Gz * y_corr - dBzdz * x_corr / 2 + Bz * offset_xp  # row remainder
        By = By + Gz * x_corr - dBzdz * y_corr / 2 + Bz * offset_yp  # row remainder
        return Bx, By, Bz


@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True)
def sum_field_particles(
    x: np.array, y: np.array, z: np.array, z_start: float, z_stop: float
) -> (np.array, np.array, np.array):
    """Sum particles fields"""
    n = int(len(x))
    Fx = np.zeros(n)
    Fy = np.zeros(n)
    Fz = np.zeros(n)
    for i in prange(n):  # pylint: disable=E1133
        if z_start <= z[i] <= z_stop:
            for j in prange(n):  # pylint: disable=E1133
                if z_start <= z[j] <= z_stop and i != j:
                    Fx[i] += (x[i] - x[j]) / ((x[j] - x[i]) ** 2 + (y[j] - y[i]) ** 2 + (z[j] - z[i]) ** 2) ** (3 / 2)
                    Fy[i] += (y[i] - y[j]) / ((x[j] - x[i]) ** 2 + (y[j] - y[i]) ** 2 + (z[j] - z[i]) ** 2) ** (3 / 2)
                    Fz[i] += (z[i] - z[j]) / ((x[j] - x[i]) ** 2 + (y[j] - y[i]) ** 2 + (z[j] - z[i]) ** 2) ** (3 / 2)
    return Fx, Fy, Fz


def get_field_beam(
    beam: Beam, acc: Accelerator, type: str, x: np.array, y: np.array, z: np.array
) -> (np.array, np.array, np.array):
    """Get an electric [MV/m] or magnetic [T] field space charge at a specific location in the accelerator"""
    assert type in ("E", "B"), "Check type field! (E or B)"
    Q = beam.total_charge
    q = Q / beam.n
    if type == "E":
        Ex, Ey, Ez = sum_field_particles(x, y, z, acc.z_start, acc.z_stop)
        return ke / (4 * np.pi) * q * Ex / 1e6, ke / (4 * np.pi) * q * Ey / 1e6, ke / (4 * np.pi) * q * Ez / 1e6
    if type == "B":
        Bx, By, Bz = sum_field_particles(x, y, z, acc.z_start, acc.z_stop)
        return km / (4 * np.pi) * q * Bx, -km / (4 * np.pi) * q * By, 0 * km / (4 * np.pi) * q * Bz


@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True)
def first_step_red(
    x: np.array,
    y: np.array,
    z: np.array,
    px: np.array,
    py: np.array,
    pz: np.array,
    Fx: np.array,
    Fy: np.array,
    Fz: np.array,
    z_start: float,
    z_stop: float,
) -> (np.array, np.array, np.array, np.array, np.array, np.array):
    """First step for Relativictic Difference Scheme"""
    n = int(len(x))
    for i in prange(n):  # pylint: disable=E1133
        if z_start <= z[i] <= z_stop:
            px[i] += 2 * Fx[i]
            py[i] += 2 * Fy[i]
            pz[i] += 2 * Fz[i]
            gamma = (1 + px[i] ** 2 + py[i] ** 2 + pz[i] ** 2) ** (1 / 2)
            vx = px[i] / gamma
            vy = py[i] / gamma
            vz = pz[i] / gamma
            x[i] += vx
            y[i] += vy
            z[i] += vz
        else:
            gamma = (1 + px[i] ** 2 + py[i] ** 2 + pz[i] ** 2) ** (1 / 2)
            vz = pz[i] / gamma
            z[i] += vz
    return x, y, z, px, py, pz


@jit(nopython=True, parallel=True, fastmath=True, cache=True, nogil=True)
def second_step_red(
    x: np.array,
    y: np.array,
    z: np.array,
    px: np.array,
    py: np.array,
    pz: np.array,
    Fx: np.array,
    Fy: np.array,
    Fz: np.array,
    z_start: float,
    z_stop: float,
) -> (np.array, np.array, np.array, np.array, np.array, np.array):
    """Second step for Relativictic Difference Scheme"""
    n = int(len(x))
    for i in prange(n):  # pylint: disable=E1133
        if z_start <= z[i] <= z_stop:
            gamma = (1 + px[i] ** 2 + py[i] ** 2 + pz[i] ** 2) ** (1 / 2)
            vx = px[i] / gamma
            vy = py[i] / gamma
            vz = pz[i] / gamma
            b2 = 1 + Fx[i] ** 2 + Fy[i] ** 2 + Fz[i] ** 2
            b1 = 2 - b2
            b3 = 2 * (vx * Fx[i] + vy * Fy[i] + vz * Fz[i])
            fx = 2 * (vy * Fz[i] - vz * Fy[i])
            fy = 2 * (vz * Fx[i] - vx * Fz[i])
            fz = 2 * (vx * Fy[i] - vy * Fx[i])
            vx = (vx * b1 + fx + Fx[i] * b3) / b2
            vy = (vy * b1 + fy + Fy[i] * b3) / b2
            vz = (vz * b1 + fz + Fz[i] * b3) / b2
            x[i] += vx
            y[i] += vy
            z[i] += vz
            px[i] = vx * gamma
            py[i] = vy * gamma
            pz[i] = vz * gamma
        else:
            gamma = (1 + px[i] ** 2 + py[i] ** 2 + pz[i] ** 2) ** (1 / 2)
            vz = pz[i] / gamma
            z[i] += vz
    return x, y, z, px, py, pz


class Simulation:
    """Simulation of the beam tracking in the accelerator"""

    def __init__(self, beam, accelerator):
        self.beam = beam
        self.acc = accelerator
        self.result = {}

    def track(self, *, n_files: int = 20) -> None:
        """Tracking!"""
        assert n_files > 0, "The number of files (n_files) must be a positive number!"

        # Init parameterss
        Y = self.beam.da
        Y[2] = Y[2] + self.acc.z_start - max(Y[2])  # set initial beam position

        z_start = self.acc.z_start
        z_stop = self.acc.z_stop
        dz = self.acc.dz
        dt = dz / c
        t_max = (z_stop - z_start) / c

        m = self.beam.type.mass
        q = self.beam.type.charge

        P0 = m * c * c / (e * 1e6)
        E0 = m * c / (q * dt * 1e6)
        B0 = m / (q * dt)

        # RED
        for t in np.arange(0, t_max, 2 * dt):
            # into the internal system
            Y[0], Y[1], Y[2] = Y[0] / dz, Y[1] / dz, Y[2] / dz
            Y[3], Y[4], Y[5] = Y[3] / P0, Y[4] / P0, Y[5] / P0

            # get electric field from accelerator
            Ex, Ey, Ez = get_field_accelerator(self.acc, "E", Y[0] * dz, Y[1] * dz, Y[2] * dz)
            Ex, Ey, Ez = Ex / E0, Ey / E0, Ez / E0

            # get electric field from beam
            if self.beam.total_charge:
                ex, ey, ez = get_field_beam(self.beam, self.acc, "E", Y[0] * dz, Y[1] * dz, Y[2] * dz)
                ex, ey, ez = ex / E0, ey / E0, ez / E0
                Ex, Ey, Ez = Ex + ex, Ey + ey, Ez + ez

            # first step RED
            Y[0], Y[1], Y[2], Y[3], Y[4], Y[5] = first_step_red(
                Y[0], Y[1], Y[2], Y[3], Y[4], Y[5], Ex, Ey, Ez, z_start / dz, z_stop / dz
            )
            gamma = np.sqrt(1 + Y[3] * Y[3] + Y[4] * Y[4] + Y[5] * Y[5])
            vz = Y[5] / gamma

            # get magnetic field from accelerator
            Bx, By, Bz = get_field_accelerator(self.acc, "B", Y[0] * dz, Y[1] * dz, Y[2] * dz)
            Bx, By, Bz = Bx / B0 / gamma, By / B0 / gamma, Bz / B0 / gamma

            # get magnetic field from beam
            if self.beam.total_charge:
                bx, by, bz = get_field_beam(self.beam, self.acc, "B", Y[0] * dz, Y[1] * dz, Y[2] * dz)
                bx, by, bz = bx * vz / B0, by * vz / B0, bz * vz / B0
                Bx, By, Bz = Bx + bx, By + by, Bz + bz

            # second step RED
            Y[0], Y[1], Y[2], Y[3], Y[4], Y[5] = second_step_red(
                Y[0], Y[1], Y[2], Y[3], Y[4], Y[5], Bx, By, Bz, z_start / dz, z_stop / dz
            )
            gamma = np.sqrt(1 + Y[3] * Y[3] + Y[4] * Y[4] + Y[5] * Y[5])
            vz = Y[5] / gamma

            # into the SI
            Y[0], Y[1], Y[2] = Y[0] * dz, Y[1] * dz, Y[2] * dz
            Y[3], Y[4], Y[5] = Y[3] * P0, Y[4] * P0, Y[5] * P0
            Bx, By, Bz = Bx * gamma * B0, By * gamma * B0, Bz * gamma * B0
            Ex, Ey, Ez = Ex * E0, Ey * E0, Ez * E0

            progress = t / t_max * 100
            meters = z_start + t * c
            X = np.row_stack((Y[0], Y[1], Y[2], Y[3], Y[4], Y[5], Bx, By, Bz, Ex, Ey, Ez))
            Xt = np.transpose(X)
            df = pd.DataFrame(Xt, columns=["x", "y", "z", "px", "py", "pz", "Bx", "By", "Bz", "Ex", "Ey", "Ez"])
            if progress % (100 // n_files) < 2 * dt / t_max * 100:
                self.result.update({round(meters, 3): df})
            print("\rz = %.2f m (%.1f %%) " % (meters, progress), end="")


Sim = Simulation



import kenv as kv
import numpy as np
import pandas as pd

from .constants import *  # pylint: disable=W0614

__all__ = ["Beam", "read_distribution_astra", "read_distribution_file"]


def read_distribution_file(fname):
    """Read distribution from file .csv"""
    df = pd.read_csv(fname, dtype="float32")
    return df["x"], df["y"], df["z"], df["px"], df["py"], df["pz"]


def read_distribution_astra(fname):
    """Read distribution from Astra file"""
    cols = ["x", "y", "z", "px", "py", "pz", "clock", "charge", "id", "flag"]
    #        m    m    m    eV/c  eV/c  eV/c  ns       nC
    df = pd.read_csv(fname, header=None, delim_whitespace=True, names=cols, dtype="float32")

    df = df[df.flag != -15]  # ignore the lost particles
    df["px"] = df["px"] / 1e6  # MeV/c
    df["py"] = df["py"] / 1e6  # MeV/c

    # remove the reference particle
    df0 = df.head(1)
    df = df.drop(df0.index)

    z0 = df0.z.values[0]
    pz0 = df0.pz.values[0]

    # Recalculate z and pz:
    df["z"] = z0 + df["clock"] * 1e-9 * c  # m
    df["pz"] = (pz0 + df["pz"]) / 1e6  # MeV/c
    return df["x"], df["y"], df["z"], df["px"], df["py"], df["pz"]


class Beam(kv.Beam):
    """Creates a beam of selected type particles"""

    def __init__(
        self,
        *,
        type: Element = electron,
        current: float = 0.0e0,
        energy: float = 0.0e0,
        radius: float = 0.0e0,
        radius_x: float = 0.0e0,
        radius_y: float = 0.0e0,
        radius_z: float = 0.0e0,
        rp: float = 0.0e0,
        radius_xp: float = 0.0e0,
        radius_yp: float = 0.0e0,
        normalized_emittance: float = 0.0e0,
        normalized_emittance_x: float = 0.0e0,
        normalized_emittance_y: float = 0.0e0,
        x: float = 0.0e0,
        y: float = 0.0e0,
        z: float = 0.0e0,
        xp: float = 0.0e0,
        yp: float = 0.0e0
    ):
        super().__init__(
            current=current,
            energy=energy,
            radius=radius,
            radius_x=radius_x,
            radius_y=radius_y,
            rp=rp,
            radius_xp=radius_xp,
            radius_yp=radius_yp,
            normalized_emittance=normalized_emittance,
            normalized_emittance_x=normalized_emittance_x,
            normalized_emittance_y=normalized_emittance_y,
            x=x,
            y=y,
            xp=xp,
            yp=yp,
        )
        self.type = type  # particles type
        self.distribution = ""
        self.n = 0.0  # quantity
        self.df = pd.DataFrame  # data frame
        self.da = np.array
        self.radius_z = radius_z
        self.z = z
        self.total_charge = (
            2 * type.charge / abs(type.charge) * self.current * self.radius_z / c / self.beta
        )  # beam charge

    def generate(self, distribution: str = "KV", n: int = 0, *, dpz: float = 0.01):
        """Beam generator

        This function generates a beam with a given distribution and initial beam displacement.
        """
        assert n > 0, "The number of particles (n) must be a positive number!"

        self.distribution = distribution
        self.n = n
        if distribution == "KV":
            s = np.random.normal(0, 1, int(self.n))
            t = np.random.normal(0, 1, int(self.n))
            u = np.random.normal(0, 1, int(self.n))  # pylint: disable=W0621
            v = np.random.normal(0, 1, int(self.n))
            norm = (s * s + t * t + u * u + v * v) ** 0.5
            (s, t, u, v) = (s, t, u, v) / norm
            (x, y) = (self.radius_x * s, self.radius_y * t)
            u = 2 * (self.px * self.normalized_emittance_x * u + self.radius_xp * x) / self.radius_x
            v = 2 * (self.py * self.normalized_emittance_y * v + self.radius_yp * y) / self.radius_y
            (x, y) = (x + self.x, y + self.y)
            (px, py) = (u + self.p * self.xp, v + self.p * self.yp)

        z = np.random.uniform(-self.radius_z + self.z, self.radius_z + self.z, int(self.n))
        pz = np.random.uniform((1 - dpz) * self.pz, (1 + dpz) * self.pz, int(self.n))

        self.da = np.row_stack((x, y, z, px, py, pz))
        self.df = pd.DataFrame(np.transpose(self.da), columns=["x", "y", "z", "px", "py", "pz"])

    def upload(self, file_name: str):
        """Particle loading"""
        file_extension = file_name.split(".")[-1]

        if file_extension == "ini":
            x, y, z, px, py, pz = read_distribution_astra(file_name)
            self.n = int(len(x))
            self.da = np.row_stack((x, y, z, px, py, pz))
            self.df = pd.DataFrame(np.transpose(self.da), columns=["x", "y", "z", "px", "py", "pz"])
        if file_extension == "csv":
            x, y, z, px, py, pz = read_distribution_file(file_name)
            self.n = int(len(x))
            self.da = np.row_stack((x, y, z, px, py, pz))
            self.df = pd.DataFrame(np.transpose(self.da), columns=["x", "y", "z", "px", "py", "pz"])

    def __str__(self):
        self.description = (
            "Beam parameters:"
            + "\n"
            + "\tType\t"
            + (self.type.name)
            + "\n"
            + "\tDistribution\t"
            + (self.distribution)
            + "\n"
            + "\tParticles\t%0.0f" % (self.n)
            + "\n"
            + "\tCurrent\t%0.0f A" % (self.current)
            + "\n"
            + "\tEnergy\t%0.3f MeV" % (self.energy)
            + "\n"
            + "\tTotal momentum\t%0.3f MeV/c" % (self.momentum)
            + "\n"
            + "\tRel. factor\t%0.3f" % (self.gamma)
            + "\n"
            + "\tRadius x\t%0.1f mm" % (self.radius_x * 1e3)
            + "\n"
            + "\tRadius y\t%0.1f mm" % (self.radius_y * 1e3)
            + "\n"
            + "\tRadius z\t%0.1f m" % (self.radius_z)
            + "\n"
            + "\tRadius x prime\t%0.1f mrad" % (self.radius_xp * 1e3)
            + "\n"
            + "\tRadius y prime\t%0.1f mrad" % (self.radius_yp * 1e3)
            + "\n"
            + "\tHorizontal centroid position\t%0.1f mm" % (self.x * 1e3)
            + "\n"
            + "\tVertical centroid position\t%0.1f mm" % (self.y * 1e3)
            + "\n"
            + "\tHorizontal centroid angle\t%0.1f mrad" % (self.xp * 1e3)
            + "\n"
            + "\tVertical centroid angle\t%0.1f mrad" % (self.yp * 1e3)
            + "\n"
            + "\tNormalized emittance x\t%0.1f mm*mrad" % (self.normalized_emittance_x * 1e6)
            + "\n"
            + "\tNormalized emittance y\t%0.1f mm*mrad" % (self.normalized_emittance_y * 1e6)
            + "\n"
        )
        return self.description



from setuptools import find_packages, setup

import redpic

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="redpic",
    version=redpic.__version__,
    author="Vyacheslav Fedorov",
    author_email="slava@fuodorov.ru",
    description=redpic.__doc__,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/fuodorov/redpic",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Physics",
    ],
)
